///////////////////////////////////////////////////////////////////////////////
// main.cpp
// ========
// drawing cubespheres using vertex array (glDrawElements)
// dependency: freeglut/glut
//
//  AUTHOR: Song Ho Ahn (song.ahn@gmail.com)
// CREATED: 2018-07-17
// UPDATED: 2024-03-13
///////////////////////////////////////////////////////////////////////////////

#include <glad/glad.h>
#include <GLFW/glfw3.h>

#include <iostream>
#include <sstream>
#include <iomanip>
#include <fstream>
#include "BitmapFontData.h"     // to draw bitmap font with GLFW
#include "fontCourier20.h"      // font:courier new, height:20px
#include "Bmp.h"
#include "Timer.h"
#include "Matrices.h"
#include "Cubesphere.h"


// glfw callbacks
void errorCallback(int error, const char* description);
void framebufferSizeCallback(GLFWwindow* window, int width, int height);
void keyCallback(GLFWwindow* window, int key, int scancode, int action, int mods);
void mouseButtonCallback(GLFWwindow* window, int button, int action, int mods);
void cursorPosCallback(GLFWwindow* window, double x, double y);

// function prototypes
void preFrame(double frameTime);
void frame(double frameTime);
void postFrame(double frameTime);
void toPerspective();
void toOrtho();
void initGL();
bool initGLSL();
void initVBO();
bool initSharedMem();
void clearSharedMem();
void showInfo();
void showFPS();
GLuint loadTexture(const char* fileName, bool wrap=true);
GLuint loadCubemap(const char* fileName);



//@@ used built-in attributes and uniforms
// blinn specular shading =================================
const char* vsSource1 = R"(
// GLSL version (OpenGL 3.3)
#version 330
// uniforms
uniform mat4 matrixModelView;
uniform mat4 matrixNormal;
uniform mat4 matrixModelViewProjection;
// vertex attribs (input)
layout(location=0) in vec3 vertexPosition;
layout(location=1) in vec3 vertexNormal;
layout(location=2) in vec2 vertexTexCoord;
// varyings (output)
out vec3 esVertex;
out vec3 esNormal;
void main()
{
    esVertex = vec3(matrixModelView * vec4(vertexPosition, 1.0));
    esNormal = vec3(matrixNormal * vec4(vertexNormal, 1.0));
    gl_Position = matrixModelViewProjection * vec4(vertexPosition, 1.0);
}
)";
const char* fsSource1 = R"(
// GLSL version (OpenGL 3.3)
#version 330
// uniforms
uniform vec4 lightPosition;             // should be in the eye space
uniform vec4 lightAmbient;              // light ambient color
uniform vec4 lightDiffuse;              // light diffuse color
uniform vec4 lightSpecular;             // light specular color
uniform vec4 materialAmbient;           // material ambient color
uniform vec4 materialDiffuse;           // material diffuse color
uniform vec4 materialSpecular;          // material specular color
uniform float materialShininess;        // material specular shininess
// varyings (input)
in vec3 esVertex;
in vec3 esNormal;
// output
out vec4 fragColor;
void main()
{
    vec3 normal = normalize(esNormal);
    vec3 view = normalize(-esVertex);
    vec3 light;
    if(lightPosition.w == 0.0)
    {
        light = normalize(lightPosition.xyz);
    }
    else
    {
        light = normalize(lightPosition.xyz - esVertex);
    }
    vec3 halfVec = normalize(light + view);
    vec4 color =  materialAmbient * lightAmbient;
    float dotNL = max(dot(normal, light), 0.0);
    color += materialDiffuse * lightDiffuse * dotNL;
    float dotNH = max(dot(normal, halfVec), 0.0);
    /*vec4 specular = (vec4(1.0) - color) * materialPpecular * lightSpecular * pow(dotNH, materialShininess);
    color += specular;*/
    color += materialSpecular * lightSpecular * pow(dotNH, materialShininess);
    fragColor = color;
}
)";



// blinn shading with texture =============================
const char* vsSource2 = R"(
// GLSL version (OpenGL 3.3)
#version 330
// uniforms
uniform mat4 matrixModelView;
uniform mat4 matrixNormal;
uniform mat4 matrixModelViewProjection;
// vertex attribs (input)
layout(location=0) in vec3 vertexPosition;
layout(location=1) in vec3 vertexNormal;
layout(location=2) in vec2 vertexTexCoord;
// varyings (output)
out vec3 esVertex;
out vec3 esNormal;
out vec2 texCoord0;
void main()
{
    esVertex = vec3(matrixModelView * vec4(vertexPosition, 1.0));
    esNormal = vec3(matrixNormal * vec4(vertexNormal, 1.0));
    texCoord0 = vertexTexCoord;
    gl_Position = matrixModelViewProjection * vec4(vertexPosition, 1.0);
}
)";

const char* fsSource2 = R"(
// GLSL version (OpenGL 3.3)
#version 330
// uniforms
uniform vec4 lightPosition;             // should be in the eye space
uniform vec4 lightAmbient;              // light ambient color
uniform vec4 lightDiffuse;              // light diffuse color
uniform vec4 lightSpecular;             // light specular color
uniform vec4 materialAmbient;           // material ambient color
uniform vec4 materialDiffuse;           // material diffuse color
uniform vec4 materialSpecular;          // material specular color
uniform float materialShininess;        // material specular shininess
uniform sampler2D map0;                 // texture map #1
// varyings (input)
in vec3 esVertex;
in vec3 esNormal;
in vec2 texCoord0;
// output
out vec4 fragColor;
void main()
{
    vec3 normal = normalize(esNormal);
    vec3 view = normalize(-esVertex);
    vec3 light;
    if(lightPosition.w == 0.0)
    {
        light = normalize(lightPosition.xyz);
    }
    else
    {
        light = normalize(lightPosition.xyz - esVertex);
    }
    vec3 halfv = normalize(light + view);
    vec4 color = materialAmbient * lightAmbient;
    float dotNL = max(dot(normal, light), 0.0);
    color += materialDiffuse * lightDiffuse * dotNL;
    color *= texture(map0, texCoord0);
    float dotNH = max(dot(normal, halfv), 0.0);
    color += materialSpecular * lightSpecular * pow(dotNH, materialShininess);
    fragColor = color;
}
)";


// blinn shading with cubemap =============================
const char* vsSource3 = R"(
// GLSL version (OpenGL 3.3)
#version 330
// uniforms
uniform mat4 matrixModelView;
uniform mat4 matrixNormal;
uniform mat4 matrixModelViewProjection;
//uniform int cubemapMode;
// vertex attribs (input)
layout(location=0) in vec3 vertexPosition;
layout(location=1) in vec3 vertexNormal;
//layout(location=2) in vec2 vertexTexCoord; // no need, generate it in fs
// varyings (output)
out vec3 esVertex;
out vec3 esNormal;
out vec3 osVertex;
void main()
{
    esVertex = vec3(matrixModelView * vec4(vertexPosition, 1.0));
    esNormal = vec3(matrixNormal * vec4(vertexNormal, 1.0));
    osVertex = vertexPosition;
    gl_Position = matrixModelViewProjection * vec4(vertexPosition, 1.0);
}
)";

const char* fsSource3 = R"(
// GLSL version (OpenGL 3.3)
#version 330
// uniforms
uniform vec4 lightPosition;             // should be in the eye space
uniform vec4 lightAmbient;              // light ambient color
uniform vec4 lightDiffuse;              // light diffuse color
uniform vec4 lightSpecular;             // light specular color
uniform vec4 materialAmbient;           // material ambient color
uniform vec4 materialDiffuse;           // material diffuse color
uniform vec4 materialSpecular;          // material specular color
uniform float materialShininess;        // material specular shininess
uniform samplerCube map0;               // cubemap #1
uniform int cubemapMode;                // cubemap mode; 1=object, 2=reflection, 3=refraction
// varyings (input)
in vec3 esVertex;
in vec3 esNormal;
in vec3 osVertex;
// output
out vec4 fragColor;
void main()
{
    vec3 normal = normalize(esNormal);
    vec3 view = normalize(-esVertex);
    vec3 light;
    if(lightPosition.w == 0.0)
    {
        light = normalize(lightPosition.xyz);
    }
    else
    {
        light = normalize(lightPosition.xyz - esVertex);
    }
    vec3 halfv = normalize(light + view);
    vec4 color = materialAmbient * lightAmbient;
    float dotNL = max(dot(normal, light), 0.0);
    color += materialDiffuse * lightDiffuse * dotNL;

    // generate texture coords
    vec3 texCoord;
    if(cubemapMode == 1)    // object-space
    {
        texCoord = osVertex;
    }
    else if(cubemapMode == 2)   // reflection
    {
        texCoord = reflect(-view, normal);
    }
    else if(cubemapMode == 3)   // refraction
    {
        texCoord = refract(view, normal, 1.0/1.52);
    }

    // apply color from cubemap
    color *= texture(map0, texCoord);

    float dotNH = max(dot(normal, halfv), 0.0);
    color += materialSpecular * lightSpecular * pow(dotNH, materialShininess);
    fragColor = color;
}
)";



// constants
const int   WINDOW_WIDTH    = 1500;
const int   WINDOW_HEIGHT   = 500;
const float CAMERA_DISTANCE = 3.7f;



// global variables
GLFWwindow* window;
int windowWidth;
int windowHeight;
int fbWidth;
int fbHeight;
double runTime;
bool mouseLeftDown;
bool mouseRightDown;
bool mouseMiddleDown;
double mouseX, mouseY;
float cameraAngleX;
float cameraAngleY;
float cameraDistance;
int drawMode;
int cubemapMode;
GLuint texId1, texId2;                  // texture object
int imageWidth;
int imageHeight;
Matrix4 matrixModelView;
Matrix4 matrixProjection;
Matrix4 matrixModelViewProjection;
Matrix4 matrixNormal;
BitmapFontData bmFont;  // bitmap font

// VBO and VAO
GLuint vboId = 0;       // ID of VBO for vertex arrays
GLuint iboId = 0;       // ID of VBO for index array
GLuint vaoId = 0;       // ID of VAO for vertex array states

// GLSL
GLuint progId1 = 0, progId2 = 0, progId3 = 0;   // IDs of GLSL programs
// uniforms
struct Uniform1
{
    GLint matrixModelView;
    GLint matrixModelViewProjection;
    GLint matrixNormal;
    GLint lightPosition;
    GLint lightAmbient;
    GLint lightDiffuse;
    GLint lightSpecular;
    GLint materialAmbient;
    GLint materialDiffuse;
    GLint materialSpecular;
    GLint materialShininess;
} uniform1;
struct Uniform2
{
    GLint matrixModelView;
    GLint matrixModelViewProjection;
    GLint matrixNormal;
    GLint lightPosition;
    GLint lightAmbient;
    GLint lightDiffuse;
    GLint lightSpecular;
    GLint materialAmbient;
    GLint materialDiffuse;
    GLint materialSpecular;
    GLint materialShininess;
    GLint map;               // tex map
} uniform2;
struct Uniform3
{
    GLint matrixModelView;
    GLint matrixModelViewProjection;
    GLint matrixNormal;
    GLint lightPosition;
    GLint lightAmbient;
    GLint lightDiffuse;
    GLint lightSpecular;
    GLint materialAmbient;
    GLint materialDiffuse;
    GLint materialSpecular;
    GLint materialShininess;
    GLint cubemap;           // cube map
    GLint cubemapMode;       // cubemap mode; 1=object, 2=reflection, 3=refraction
} uniform3;
GLint uniformCubemapMode;       // cubemap mode; 1=object, 2=reflection, 3=refraction
GLint attribVertexPosition = 0; // 0
GLint attribVertexNormal = 1;   // 1
GLint attribVertexTexCoord = 2; // 2

// for cubed sphere
int subdivision = 3;
//Cubesphere sphere;                          // create default sphere(1, 3, true)
Cubesphere sphere(1.0f, subdivision, false); // radius, subdivision, smooth



///////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    // init global vars
    initSharedMem();

    // init GLFW
    glfwInit();
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    //glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_COMPAT_PROFILE);
    //NOTE: mac only supports forward-compatible, core profile for v3.x & v4.x
    #ifdef __APPLE__
        glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
    #endif

    window = glfwCreateWindow(WINDOW_WIDTH, WINDOW_HEIGHT, argv[0], 0, 0);
    if(window)
    {
        int ww, wh, fw, fh;
        glfwGetWindowSize(window, &ww, &wh);
        glfwGetFramebufferSize(window, &fw, &fh);
        std::cout << "GLFW window is created: " << ww << "x" << wh
                  << " (Framebuffer: " << fw << "x" << fh << ")" << std::endl;
    }
    else
    {
        std::cout << "[ERROR] Failed to create GLFW window" << std::endl;
        glfwTerminate();
        exit(1);
    }

    // set OpenGL RC
    glfwMakeContextCurrent(window);

    // init glad to load gl extensions
    gladLoadGLLoader((GLADloadproc)glfwGetProcAddress);

    //DEBUG: print glfw/opengl info
    //int majorVersion = 0, minorVersion = 0, revision = 0;
    //glfwGetVersion(&majorVersion, &minorVersion, &revision);
    //std::cout << "OpenGL Version: " << majorVersion << "." << minorVersion << "." << revision << std::endl;
    std::cout << "OpenGL Version: " << glGetString(GL_VERSION) << std::endl;
    std::cout << "GLSL Version: " << glGetString(GL_SHADING_LANGUAGE_VERSION) << std::endl;

    // init OpenGL, vbo glsl after setting RC
    initGL();
    initGLSL();
    initVBO();

    // load bitmap font from array, call it after GL/GLSL is initialized
    bmFont.loadFont(fontCourier20, bitmapCourier20);

    // init GLFW callbacks
    glfwSetFramebufferSizeCallback(window, framebufferSizeCallback);
    glfwSetKeyCallback(window, keyCallback);
    glfwSetMouseButtonCallback(window, mouseButtonCallback);
    glfwSetCursorPosCallback(window, cursorPosCallback);
    glfwSetErrorCallback(errorCallback);

    // load 2D texture
    texId1 = loadTexture("grid512.bmp", true);

    // load cube map
    texId2 = loadCubemap("cubemap.bmp");

    // init projection matrix
    toPerspective();

    // main rendering loop
    glfwSwapInterval(1); // frequency
    while(!glfwWindowShouldClose(window))
    {
        // get frame time
        double currTime = glfwGetTime();
        double frameTime = currTime - runTime;
        runTime = currTime;

        // draw
        preFrame(frameTime);
        frame(frameTime);
        postFrame(frameTime);

        // get events
        glfwPollEvents();
    }

    // terminate program
    glfwTerminate();
    std::cout << "Terminate program" << std::endl;
    return 0;
}



///////////////////////////////////////////////////////////////////////////////
// initialize OpenGL
// disable unused features
///////////////////////////////////////////////////////////////////////////////
void initGL()
{
    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);      // 4-byte pixel alignment

    // enable /disable features
    //glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    //glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    //glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);

    glClearColor(0, 0, 0, 0);                   // background color
    glClearStencil(0);                          // clear stencil buffer
    glClearDepth(1.0f);                         // 0 is near, 1 is far
    glDepthFunc(GL_LEQUAL);
}



///////////////////////////////////////////////////////////////////////////////
// create glsl programs
///////////////////////////////////////////////////////////////////////////////
bool initGLSL()
{
    const int MAX_LENGTH = 2048;
    char log[MAX_LENGTH];
    int logLength = 0;
    int vsStatus, fsStatus;
    bool result;

    // light and material colors
    float lightPosition[] = {0, 0, 1, 0};
    float lightAmbient[]  = {0.5f, 0.5f, 0.5f, 1};
    float lightDiffuse[]  = {0.7f, 0.7f, 0.7f, 1};
    float lightSpecular[] = {1.0f, 1.0f, 1.0f, 1};
    float materialAmbient[]  = {0.5f, 0.5f, 0.5f, 1};
    float materialDiffuse[]  = {0.9f, 0.9f, 0.9f, 1};
    float materialSpecular[] = {0.7f, 0.7f, 0.7f, 1};
    float materialShininess  = 128;

    // create shader and program: blinn shader ================================
    GLuint vsId1 = glCreateShader(GL_VERTEX_SHADER);
    GLuint fsId1 = glCreateShader(GL_FRAGMENT_SHADER);
    progId1 = glCreateProgram();

    // load shader sources
    glShaderSource(vsId1, 1, &vsSource1, NULL);
    glShaderSource(fsId1, 1, &fsSource1, NULL);

    // compile shader sources
    glCompileShader(vsId1);
    glCompileShader(fsId1);

    // attach shaders to the program and link
    glAttachShader(progId1, vsId1);
    glAttachShader(progId1, fsId1);
    glLinkProgram(progId1);

    // set uniform values
    glUseProgram(progId1);
    uniform1.matrixModelView           = glGetUniformLocation(progId1, "matrixModelView");
    uniform1.matrixModelViewProjection = glGetUniformLocation(progId1, "matrixModelViewProjection");
    uniform1.matrixNormal              = glGetUniformLocation(progId1, "matrixNormal");
    uniform1.lightPosition             = glGetUniformLocation(progId1, "lightPosition");
    uniform1.lightAmbient              = glGetUniformLocation(progId1, "lightAmbient");
    uniform1.lightDiffuse              = glGetUniformLocation(progId1, "lightDiffuse");
    uniform1.lightSpecular             = glGetUniformLocation(progId1, "lightSpecular");
    uniform1.materialAmbient           = glGetUniformLocation(progId1, "materialAmbient");
    uniform1.materialDiffuse           = glGetUniformLocation(progId1, "materialDiffuse");
    uniform1.materialSpecular          = glGetUniformLocation(progId1, "materialSpecular");
    uniform1.materialShininess         = glGetUniformLocation(progId1, "materialShininess");
    glUniform4fv(uniform1.lightPosition, 1, lightPosition);
    glUniform4fv(uniform1.lightAmbient, 1, lightAmbient);
    glUniform4fv(uniform1.lightDiffuse, 1, lightDiffuse);
    glUniform4fv(uniform1.lightSpecular, 1, lightSpecular);
    glUniform4fv(uniform1.materialAmbient, 1, materialAmbient);
    glUniform4fv(uniform1.materialDiffuse, 1, materialDiffuse);
    glUniform4fv(uniform1.materialSpecular, 1, materialSpecular);
    glUniform1f(uniform1.materialShininess, materialShininess);

    //@@ debug
    glGetShaderiv(vsId1, GL_COMPILE_STATUS, &vsStatus);
    if(vsStatus == GL_FALSE)
    {
        glGetShaderiv(vsId1, GL_INFO_LOG_LENGTH, &logLength);
        glGetShaderInfoLog(vsId1, MAX_LENGTH, &logLength, log);
        std::cout << "===== Vertex Shader Log #1 =====\n" << log << std::endl;
    }
    glGetShaderiv(fsId1, GL_COMPILE_STATUS, &fsStatus);
    if(fsStatus == GL_FALSE)
    {
        glGetShaderiv(fsId1, GL_INFO_LOG_LENGTH, &logLength);
        glGetShaderInfoLog(fsId1, MAX_LENGTH, &logLength, log);
        std::cout << "===== Fragment Shader Log #1 =====\n" << log << std::endl;
    }

    // create second shader and program: blinn + 2D texture ===================
    GLuint vsId2 = glCreateShader(GL_VERTEX_SHADER);
    GLuint fsId2 = glCreateShader(GL_FRAGMENT_SHADER);
    progId2 = glCreateProgram();
    glShaderSource(vsId2, 1, &vsSource2, NULL);
    glShaderSource(fsId2, 1, &fsSource2, NULL);
    glCompileShader(vsId2);
    glCompileShader(fsId2);

    // attach shaders to the program and link
    glAttachShader(progId2, vsId2);
    glAttachShader(progId2, fsId2);
    glLinkProgram(progId2);

    // set uniform values
    glUseProgram(progId2);
    uniform2.matrixModelView           = glGetUniformLocation(progId2, "matrixModelView");
    uniform2.matrixModelViewProjection = glGetUniformLocation(progId2, "matrixModelViewProjection");
    uniform2.matrixNormal              = glGetUniformLocation(progId2, "matrixNormal");
    uniform2.lightPosition             = glGetUniformLocation(progId2, "lightPosition");
    uniform2.lightAmbient              = glGetUniformLocation(progId2, "lightAmbient");
    uniform2.lightDiffuse              = glGetUniformLocation(progId2, "lightDiffuse");
    uniform2.lightSpecular             = glGetUniformLocation(progId2, "lightSpecular");
    uniform2.materialAmbient           = glGetUniformLocation(progId2, "materialAmbient");
    uniform2.materialDiffuse           = glGetUniformLocation(progId2, "materialDiffuse");
    uniform2.materialSpecular          = glGetUniformLocation(progId2, "materialSpecular");
    uniform2.materialShininess         = glGetUniformLocation(progId2, "materialShininess");
    uniform2.map                       = glGetUniformLocation(progId2, "map0");
    glUniform4fv(uniform2.lightPosition, 1, lightPosition);
    glUniform4fv(uniform2.lightAmbient, 1, lightAmbient);
    glUniform4fv(uniform2.lightDiffuse, 1, lightDiffuse);
    glUniform4fv(uniform2.lightSpecular, 1, lightSpecular);
    glUniform4fv(uniform2.materialAmbient, 1, materialAmbient);
    glUniform4fv(uniform2.materialDiffuse, 1, materialDiffuse);
    glUniform4fv(uniform2.materialSpecular, 1, materialSpecular);
    glUniform1f(uniform2.materialShininess, materialShininess);
    glUniform1i(uniform2.map, 0);

    //@@ debug
    glGetShaderiv(vsId2, GL_COMPILE_STATUS, &vsStatus);
    if(vsStatus == GL_FALSE)
    {
        glGetShaderiv(vsId2, GL_INFO_LOG_LENGTH, &logLength);
        glGetShaderInfoLog(vsId2, MAX_LENGTH, &logLength, log);
        std::cout << "===== Vertex Shader Log #2 =====\n" << log << std::endl;
    }
    glGetShaderiv(fsId2, GL_COMPILE_STATUS, &fsStatus);
    if(fsStatus == GL_FALSE)
    {
        glGetShaderiv(fsId2, GL_INFO_LOG_LENGTH, &logLength);
        glGetShaderInfoLog(fsId2, MAX_LENGTH, &logLength, log);
        std::cout << "===== Fragment Shader Log #2 =====\n" << log << std::endl;
    }

    // create second shader and program: blinn + cubemap ======================
    GLuint vsId3 = glCreateShader(GL_VERTEX_SHADER);
    GLuint fsId3 = glCreateShader(GL_FRAGMENT_SHADER);
    progId3 = glCreateProgram();
    glShaderSource(vsId3, 1, &vsSource3, NULL);
    glShaderSource(fsId3, 1, &fsSource3, NULL);
    glCompileShader(vsId3);
    glCompileShader(fsId3);

    // attach shaders to the program and link
    glAttachShader(progId3, vsId3);
    glAttachShader(progId3, fsId3);
    glLinkProgram(progId3);

    // set uniform locations
    glUseProgram(progId3);
    uniform3.matrixModelView           = glGetUniformLocation(progId3, "matrixModelView");
    uniform3.matrixModelViewProjection = glGetUniformLocation(progId3, "matrixModelViewProjection");
    uniform3.matrixNormal              = glGetUniformLocation(progId3, "matrixNormal");
    uniform3.lightPosition             = glGetUniformLocation(progId3, "lightPosition");
    uniform3.lightAmbient              = glGetUniformLocation(progId3, "lightAmbient");
    uniform3.lightDiffuse              = glGetUniformLocation(progId3, "lightDiffuse");
    uniform3.lightSpecular             = glGetUniformLocation(progId3, "lightSpecular");
    uniform3.materialAmbient           = glGetUniformLocation(progId3, "materialAmbient");
    uniform3.materialDiffuse           = glGetUniformLocation(progId3, "materialDiffuse");
    uniform3.materialSpecular          = glGetUniformLocation(progId3, "materialSpecular");
    uniform3.materialShininess         = glGetUniformLocation(progId3, "materialShininess");
    uniform3.cubemap                   = glGetUniformLocation(progId3, "map0");
    uniform3.cubemapMode               = glGetUniformLocation(progId3, "cubemapMode");
    glUniform4fv(uniform3.lightPosition, 1, lightPosition);
    glUniform4fv(uniform3.lightAmbient, 1, lightAmbient);
    glUniform4fv(uniform3.lightDiffuse, 1, lightDiffuse);
    glUniform4fv(uniform3.lightSpecular, 1, lightSpecular);
    glUniform4fv(uniform3.materialAmbient, 1, materialAmbient);
    glUniform4fv(uniform3.materialDiffuse, 1, materialDiffuse);
    glUniform4fv(uniform3.materialSpecular, 1, materialSpecular);
    glUniform1f(uniform3.materialShininess, materialShininess);
    glUniform1i(uniform3.cubemap, 0);
    glUniform1i(uniform3.cubemapMode, cubemapMode);

    //@@ debug
    glGetShaderiv(vsId3, GL_COMPILE_STATUS, &vsStatus);
    if(vsStatus == GL_FALSE)
    {
        glGetShaderiv(vsId3, GL_INFO_LOG_LENGTH, &logLength);
        glGetShaderInfoLog(vsId3, MAX_LENGTH, &logLength, log);
        std::cout << "===== Vertex Shader Log #3 =====\n" << log << std::endl;
    }
    glGetShaderiv(fsId3, GL_COMPILE_STATUS, &fsStatus);
    if(fsStatus == GL_FALSE)
    {
        glGetShaderiv(fsId3, GL_INFO_LOG_LENGTH, &logLength);
        glGetShaderInfoLog(fsId3, MAX_LENGTH, &logLength, log);
        std::cout << "===== Fragment Shader Log #3 =====\n" << log << std::endl;
    }

    // unbind GLSL
    glUseProgram(0);

    // check GLSL status
    int linkStatus1, linkStatus2, linkStatus3;
    glGetProgramiv(progId1, GL_LINK_STATUS, &linkStatus1);
    glGetProgramiv(progId2, GL_LINK_STATUS, &linkStatus2);
    glGetProgramiv(progId3, GL_LINK_STATUS, &linkStatus3);
    if(linkStatus1 == GL_FALSE)
    {
        glGetProgramiv(progId1, GL_INFO_LOG_LENGTH, &logLength);
        glGetProgramInfoLog(progId1, MAX_LENGTH, &logLength, log);
        std::cout << "===== GLSL Program Log #1 =====\n" << log << std::endl;
        result = false;
    }
    else if(linkStatus2 == GL_FALSE)
    {
        glGetProgramiv(progId2, GL_INFO_LOG_LENGTH, &logLength);
        glGetProgramInfoLog(progId2, MAX_LENGTH, &logLength, log);
        std::cout << "===== GLSL Program Log #2 =====\n" << log << std::endl;
        result = false;
    }
    else if(linkStatus3 == GL_FALSE)
    {
        glGetProgramiv(progId3, GL_INFO_LOG_LENGTH, &logLength);
        glGetProgramInfoLog(progId3, MAX_LENGTH, &logLength, log);
        std::cout << "===== GLSL Program Log #2 =====\n" << log << std::endl;
        result = false;
    }
    else
    {
        result = true;
    }

    return result;
}



///////////////////////////////////////////////////////////////////////////////
// copy vertex data to VBO and VA state to VAO
///////////////////////////////////////////////////////////////////////////////
void initVBO()
{
    // create vertex array object to store all vertex array states only once
    if(!vaoId)
        glGenVertexArrays(1, &vaoId);
    glBindVertexArray(vaoId);

    // create vertex buffer objects
    if(!vboId)
        glGenBuffers(1, &vboId);

    glBindBuffer(GL_ARRAY_BUFFER, vboId);
    glBufferData(GL_ARRAY_BUFFER, sphere.getInterleavedVertexSize(), sphere.getInterleavedVertices(), GL_STATIC_DRAW);

    if(!iboId)
        glGenBuffers(1, &iboId);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, iboId);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sphere.getIndexSize(), sphere.getIndices(), GL_STATIC_DRAW);

    // enable vertex array attributes for bound VAO
    glEnableVertexAttribArray(attribVertexPosition);
    glEnableVertexAttribArray(attribVertexNormal);
    glEnableVertexAttribArray(attribVertexTexCoord);

    // store vertex array pointers to bound VAO
    int stride = sphere.getInterleavedStride();
    glVertexAttribPointer(attribVertexPosition, 3, GL_FLOAT, false, stride, 0);
    glVertexAttribPointer(attribVertexNormal, 3, GL_FLOAT, false, stride, (void*)(3 * sizeof(float)));
    glVertexAttribPointer(attribVertexTexCoord, 2, GL_FLOAT, false, stride, (void*)(6 * sizeof(float)));

    // unbind
    glBindVertexArray(0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}



///////////////////////////////////////////////////////////////////////////////
// initialize global variables
///////////////////////////////////////////////////////////////////////////////
bool initSharedMem()
{
    windowWidth = fbWidth = WINDOW_WIDTH;
    windowHeight = fbHeight = WINDOW_HEIGHT;

    runTime = 0;

    mouseLeftDown = mouseRightDown = mouseMiddleDown = false;
    mouseX = mouseY = 0;

    cameraAngleX = cameraAngleY = 0.0f;
    cameraDistance = CAMERA_DISTANCE;

    drawMode = 0;       // 0:fill, 1: wireframe, 2:points
    cubemapMode = 1;    // 1:object, 2:reflection, 3:refraction

    //DEBUG
    sphere.printSelf();
    std::cout << std::endl;

    return true;
}



///////////////////////////////////////////////////////////////////////////////
// clean up global vars
///////////////////////////////////////////////////////////////////////////////
void clearSharedMem()
{
    // clean up VBOs
    glDeleteBuffers(1, &vboId);
    glDeleteBuffers(1, &iboId);
    vboId = iboId = 0;

    // clean up VAOs
    glDeleteVertexArrays(1, &vaoId);
    vaoId = 0;

    // clean up tex
    glDeleteTextures(1, &texId1);
    glDeleteTextures(1, &texId2);
    texId1 = texId2 = 0;
}



///////////////////////////////////////////////////////////////////////////////
// load raw image as a texture
///////////////////////////////////////////////////////////////////////////////
GLuint loadTexture(const char* fileName, bool wrap)
{
    Image::Bmp bmp;
    if(!bmp.read(fileName))
        return 0;     // exit if failed load image

    // get bmp info
    int width = bmp.getWidth();
    int height = bmp.getHeight();
    const unsigned char* data = bmp.getDataRGB();
    GLenum type = GL_UNSIGNED_BYTE;    // only allow BMP with 8-bit per channel

    // We assume the image is 8-bit, 24-bit or 32-bit BMP
    GLenum format;
    int bpp = bmp.getBitCount();
    if(bpp == 8)
        format = GL_LUMINANCE;
    else if(bpp == 24)
        format = GL_RGB;
    else if(bpp == 32)
        format = GL_RGBA;
    else
        return 0;               // NOT supported, exit

    // gen texture ID
    GLuint texture;
    glGenTextures(1, &texture);

    // set active texture and configure it
    glBindTexture(GL_TEXTURE_2D, texture);

    // select modulate to mix texture with color for shading (default mode)
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    //glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE); // removed in v3.1

    // if wrap is true, the texture wraps over at the edges (repeat)
    //       ... false, the texture ends at the edges (clamp)
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap ? GL_REPEAT : GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap ? GL_REPEAT : GL_CLAMP);
    //glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    //glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    // copy texture data
    glTexImage2D(GL_TEXTURE_2D, 0, format, width, height, 0, format, type, data);
    glGenerateMipmap(GL_TEXTURE_2D);

    /*
    // build our texture mipmaps
    switch(bpp)
    {
    case 8:
        gluBuild2DMipmaps(GL_TEXTURE_2D, 1, width, height, GL_LUMINANCE, type, data);
        break;
    case 24:
        gluBuild2DMipmaps(GL_TEXTURE_2D, 3, width, height, GL_RGB, type, data);
        break;
    case 32:
        gluBuild2DMipmaps(GL_TEXTURE_2D, 4, width, height, GL_RGBA, type, data);
        break;
    }
    */

    return texture;
}



///////////////////////////////////////////////////////////////////////////////
// load 6 cubemap textures by inserting number 0...6 right after the input file
// For example, if input is "image.bmp", it will load the following image files
// image0.bmp: GL_TEXTURE_CUBE_MAP_POSITIVE_X
// image1.bmp: GL_TEXTURE_CUBE_MAP_NEGATIVE_X
// image2.bmp: GL_TEXTURE_CUBE_MAP_POSITIVE_Y
// image3.bmp: GL_TEXTURE_CUBE_MAP_NEGATIVE_Y
// image4.bmp: GL_TEXTURE_CUBE_MAP_POSITIVE_Z
// image5.bmp: GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
///////////////////////////////////////////////////////////////////////////////
GLuint loadCubemap(const char* fileStr)
{
    std::string inputFile = fileStr;
    std::size_t pos = inputFile.find_last_of('.');
    if(pos == std::string::npos)
    {
        std::cout << "Incorrect file name (no extension)" << std::endl;
        return 0;
    }

    std::string fileName = inputFile.substr(0, pos);
    std::string fileExt = inputFile.substr(pos);    // with .
    //std::cout << fileName << fileExt << std::endl;

    // gen texture ID
    GLuint texture;
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_CUBE_MAP, texture);

    // set texture parameters
    //glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, GL_TRUE); // removed in v3.1
    glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    //glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

    // load 6 bitmaps by appending 0...6 in the file names
    for(int i = 0; i < 6; ++i)
    {
        std::string file = fileName + std::to_string(i) + fileExt;
        Image::Bmp bmp;
        if(!bmp.read(file.c_str()))
        {
            std::cout << "Failed to open " << file << std::endl;
            return 0;     // exit if failed load image
        }

        // get bmp info
        int width = bmp.getWidth();
        int height = bmp.getHeight();
        const unsigned char* data = bmp.getDataRGB();
        GLenum type = GL_UNSIGNED_BYTE;    // only allow BMP with 8-bit per channel
        //bmp.printSelf();

        // We assume the image is 8-bit, 24-bit or 32-bit BMP
        GLenum format;
        int bpp = bmp.getBitCount();
        if(bpp == 8)
            format = GL_LUMINANCE;
        else if(bpp == 24)
            format = GL_RGB;
        else if(bpp == 32)
            format = GL_RGBA;
        else
        {
            std::cout << "Not supported bitmap format" << std::endl;
            return 0;   // NOT supported, exit
        }

        // copy bitmap to each face of cubemap
        glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X+i,  // target (0~5)
                     0,                                 // level
                     format,                            // internal format
                     width,                             // image width
                     height,                            // image height
                     0,                                 // border
                     format,                            // format
                     type,                              // type
                     data);                             // bitmap data
        // must generate mipmaps explicitly after copying bitmap
        glGenerateMipmap(GL_TEXTURE_CUBE_MAP);
    }

    // unbind
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    return texture;
}



///////////////////////////////////////////////////////////////////////////////
// display info messages
///////////////////////////////////////////////////////////////////////////////
void showInfo()
{
    toOrtho();

    // call it once before drawing text to configure orthographic projection
    bmFont.setWindowSize(windowWidth, windowHeight);

    std::stringstream ss;
    ss << std::fixed << std::setprecision(3);

    int x = 1;
    int y = windowHeight - bmFont.getBaseline();
    bmFont.setColor(1, 1, 1, 1);

    ss << "Subdivision: " << sphere.getSubdivision() << std::ends;
    bmFont.drawText(x, y, ss.str().c_str());
    ss.str("");
    y -= bmFont.getHeight();

    ss << "Radius: " << sphere.getRadius() << std::ends;
    bmFont.drawText(x, y, ss.str().c_str());
    ss.str("");
    y -= bmFont.getHeight();

    ss << "Triangle Count: " << sphere.getTriangleCount() << std::ends;
    bmFont.drawText(x, y, ss.str().c_str());
    ss.str("");
    y -= bmFont.getHeight();

    ss << "Vertex Count: " << sphere.getVertexCount() << std::ends;
    bmFont.drawText(x, y, ss.str().c_str());
    ss.str("");
    y -= bmFont.getHeight();

    ss << "Index Count: " << sphere.getIndexCount() << std::ends;
    bmFont.drawText(x, y, ss.str().c_str());
    ss.str("");
    y -= bmFont.getHeight();

    bmFont.drawText(x, 1, "Press SPACE to change subdivision level");

    // unset floating format
    ss << std::resetiosflags(std::ios_base::fixed | std::ios_base::floatfield);

    showFPS();

    // go back to perspective mode
    toPerspective();
}



///////////////////////////////////////////////////////////////////////////////
// display frame rates
///////////////////////////////////////////////////////////////////////////////
void showFPS()
{
    static Timer timer;
    static int count = 0;
    static std::string fps = "0.0 FPS";
    double elapsedTime = 0.0;

    ++count;

    // update fps every second
    elapsedTime = timer.getElapsedTime();
    if(elapsedTime >= 1.0)
    {
        std::stringstream ss;
        ss << std::fixed << std::setprecision(1);
        ss << (count / elapsedTime) << " FPS" << std::ends; // update fps string
        ss << std::resetiosflags(std::ios_base::fixed | std::ios_base::floatfield);
        fps = ss.str();
        count = 0;                      // reset counter
        timer.start();                  // restart timer
    }

    // draw FPS at top-right corner
    int x = windowWidth - bmFont.getTextWidth(fps.c_str()) - 1;
    int y = windowHeight - bmFont.getBaseline();
    bmFont.setColor(1, 1, 0, 1.0f);
    bmFont.drawText(x, y, fps.c_str());
}



///////////////////////////////////////////////////////////////////////////////
// set projection matrix as orthogonal
///////////////////////////////////////////////////////////////////////////////
void toOrtho()
{
    const float N = -1.0f;
    const float F = 1.0f;

    // get current dimensions
    glfwGetWindowSize(window, &windowWidth, &windowHeight); // get window size
    glfwGetFramebufferSize(window, &fbWidth, &fbHeight);    // get framebuffer size

    // set viewport to be the entire framebuffer size
    glViewport(0, 0, (GLsizei)fbWidth, (GLsizei)fbHeight);

    // construct ortho projection matrix, not framebuffer size
    matrixProjection.identity();
    matrixProjection[0]  =  2.0f / windowWidth;
    matrixProjection[5]  =  2.0f / windowHeight;
    matrixProjection[10] = -2.0f / (F - N);
    matrixProjection[12] = -1.0f;
    matrixProjection[13] = -1.0f;
    matrixProjection[14] = -(F + N) / (F - N);
}



///////////////////////////////////////////////////////////////////////////////
// set the projection matrix as perspective
///////////////////////////////////////////////////////////////////////////////
void toPerspective()
{
    const float N = 0.1f;
    const float F = 100.0f;
    const float PI = acos(-1.0f);
    const float FOV_Y = 40.0f / 180.0f * PI;    // in radian

    // get current dimensions
    glfwGetWindowSize(window, &windowWidth, &windowHeight); // get window size
    glfwGetFramebufferSize(window, &fbWidth, &fbHeight);    // get framebuffer size

    // set viewport to be the entire framebuffer size
    glViewport(0, 0, (GLsizei)fbWidth, (GLsizei)fbHeight);

    // construct perspective projection matrix
    float aspectRatio = (float)(windowWidth) / windowHeight;
    float tangent = tanf(FOV_Y / 2.0f);     // tangent of half fovY
    float h = N * tangent;                  // half height of near plane
    float w = h * aspectRatio;              // half width of near plane
    matrixProjection.identity();
    matrixProjection[0]  =  N / w;
    matrixProjection[5]  =  N / h;
    matrixProjection[10] = -(F + N) / (F - N);
    matrixProjection[11] = -1;
    matrixProjection[14] = -(2 * F * N) / (F - N);
    matrixProjection[15] =  0;
}







///////////////////////////////////////////////////////////////////////////////
// render each frame
///////////////////////////////////////////////////////////////////////////////
void preFrame(double frameTime)
{
}

void frame(double frameTime)
{
    // clear buffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    // transform camera (view)
    Matrix4 matrixView;
    matrixView.translate(0, 0, -cameraDistance);

    // common model matrix
    Matrix4 matrixModelCommon;
    matrixModelCommon.rotateY(cameraAngleY);
    matrixModelCommon.rotateX(cameraAngleX);

    // model matrix for each instance
    Matrix4 matrixModel1(matrixModelCommon);    // left
    Matrix4 matrixModel2(matrixModelCommon);    // center
    Matrix4 matrixModel3(matrixModelCommon);    // right
    matrixModel1.translate(-2.5f, 0, 0);        // shift left
    matrixModel3.translate(2.5f, 0, 0);         // shift right

    // bind vao
    glBindVertexArray(vaoId);

    // draw left sphere without texture
    glUseProgram(progId1);
    matrixModelView = matrixView * matrixModel1;
    matrixModelViewProjection = matrixProjection * matrixModelView;
    matrixNormal = matrixModelView;
    matrixNormal.setColumn(3, Vector4(0,0,0,1));
    glUniformMatrix4fv(uniform1.matrixModelView, 1, false, matrixModelView.get());
    glUniformMatrix4fv(uniform1.matrixModelViewProjection, 1, false, matrixModelViewProjection.get());
    glUniformMatrix4fv(uniform1.matrixNormal, 1, false, matrixNormal.get());
    glDrawElements(GL_TRIANGLES,            // primitive type
                   sphere.getIndexCount(),  // # of indices
                   GL_UNSIGNED_INT,         // data type
                   (void*)0);               // ptr to indices

    // draw center sphere with 2d texture
    glUseProgram(progId2);

    glBindTexture(GL_TEXTURE_2D, texId1);
    matrixModelView = matrixView * matrixModel2;
    matrixModelViewProjection = matrixProjection * matrixModelView;
    matrixNormal = matrixModelView;
    matrixNormal.setColumn(3, Vector4(0,0,0,1));
    glUniformMatrix4fv(uniform2.matrixModelView, 1, false, matrixModelView.get());
    glUniformMatrix4fv(uniform2.matrixModelViewProjection, 1, false, matrixModelViewProjection.get());
    glUniformMatrix4fv(uniform2.matrixNormal, 1, false, matrixNormal.get());
    glDrawElements(GL_TRIANGLES,            // primitive type
                   sphere.getIndexCount(),  // # of indices
                   GL_UNSIGNED_INT,         // data type
                   (void*)0);               // ptr to indices

    // draw right sphere with cubemap
    glUseProgram(progId3);
    glBindTexture(GL_TEXTURE_CUBE_MAP, texId2);
    matrixModelView = matrixView * matrixModel3;
    matrixModelViewProjection = matrixProjection * matrixModelView;
    matrixNormal = matrixModelView;
    matrixNormal.setColumn(3, Vector4(0,0,0,1));
    glUniformMatrix4fv(uniform3.matrixModelView, 1, false, matrixModelView.get());
    glUniformMatrix4fv(uniform3.matrixModelViewProjection, 1, false, matrixModelViewProjection.get());
    glUniformMatrix4fv(uniform3.matrixNormal, 1, false, matrixNormal.get());
    glUniform1i(uniform3.cubemapMode, cubemapMode);
    glDrawElements(GL_TRIANGLES,            // primitive type
                   sphere.getIndexCount(),  // # of indices
                   GL_UNSIGNED_INT,         // data type
                   (void*)0);               // ptr to indices

    // unbind
    glBindTexture(GL_TEXTURE_2D, 0);
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    glBindVertexArray(0);
    glUseProgram(0);

    showInfo();

    glPopMatrix();

    glfwSwapBuffers(window);
}

void postFrame(double frameTime)
{
    static double elapsedTime = 0.0;
    static int frameCount = 0;
    elapsedTime += frameTime;
    ++frameCount;
    if(elapsedTime > 1.0)
    {
        double fps = frameCount / elapsedTime;
        elapsedTime = 0;
        frameCount = 0;
        //std::cout << "FPS: " << fps << std::endl;
    }
}




//=============================================================================
// GLFW CALLBACKS
//=============================================================================
void errorCallback(int error, const char* description)
{
    std::cout << "[ERROR]: " << description << std::endl;
}

void framebufferSizeCallback(GLFWwindow* window, int w, int h)
{
    toPerspective();
    std::cout << "Framebuffer resized: " << fbWidth << "x" << fbHeight
              << " (Window: " << windowWidth << "x" << windowHeight << ")" << std::endl;
}

void keyCallback(GLFWwindow* window, int key, int scancode, int action, int mods)
{
    if(key == GLFW_KEY_ESCAPE && action == GLFW_PRESS)
    {
        //std::cout << "Closing program" << std::endl;
        clearSharedMem();
        glfwSetWindowShouldClose(window, GLFW_TRUE);
    }
    else if(key == GLFW_KEY_SPACE && action == GLFW_PRESS)
    {
        ++subdivision;
        subdivision %= 7; // from 0 to 6
        sphere.setSubdivision(subdivision);
        initVBO();
    }
    else if(key == GLFW_KEY_D && action == GLFW_PRESS)
    {
        ++drawMode;
        drawMode %= 3;
        if(drawMode == 0)        // fill mode
        {
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            glEnable(GL_DEPTH_TEST);
            glEnable(GL_CULL_FACE);
        }
        else if(drawMode == 1)  // wireframe mode
        {
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            glDisable(GL_DEPTH_TEST);
            glDisable(GL_CULL_FACE);
        }
        else                    // point mode
        {
            glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
            glDisable(GL_DEPTH_TEST);
            glDisable(GL_CULL_FACE);
        }
    }
    else if(key == GLFW_KEY_M && action == GLFW_PRESS)
    {
        ++cubemapMode;
        if(cubemapMode > 3)
            cubemapMode = 1;
    }
}

void mouseButtonCallback(GLFWwindow* window, int button, int action, int mods)
{
    // remember mouse position
    glfwGetCursorPos(window, &mouseX, &mouseY);

    if(button == GLFW_MOUSE_BUTTON_LEFT)
    {
        if(action == GLFW_PRESS)
            mouseLeftDown = true;
        else if(action == GLFW_RELEASE)
            mouseLeftDown = false;
    }
    else if(button == GLFW_MOUSE_BUTTON_RIGHT)
    {
        if(action == GLFW_PRESS)
            mouseRightDown = true;
        else if(action == GLFW_RELEASE)
            mouseRightDown = false;
    }
    else if(button == GLFW_MOUSE_BUTTON_MIDDLE)
    {
        if(action == GLFW_PRESS)
            mouseMiddleDown = true;
        else if(action == GLFW_RELEASE)
            mouseMiddleDown = false;
    }
}

void cursorPosCallback(GLFWwindow* window, double x, double y)
{
    if(mouseLeftDown)
    {
        cameraAngleY += (x - mouseX);
        cameraAngleX += (y - mouseY);
        mouseX = x;
        mouseY = y;
    }
    if(mouseRightDown)
    {
        cameraDistance -= (y - mouseY) * 0.2f;
        mouseY = y;
    }
}
